#################################################################################
# Created       : Martin Foster
# Modified      : 11-Jul-2003
#################################################################################
#
# Login - Modules part of Ethereal Realms, designed to unify authenthication 
# Copyright (C) 2000-2003  Martin Foster
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
# Author of this script can be contacted at the following:
#       E-Mail  : martin@ethereal-realms.org
#       Address : 204 - 817, 5th Street NE
#                 Calgary, Alberta
#                 T2E 3W9
#
#################################################################################

package eRealms::Login;						# Package name

use eRealms::Audit;						# Audit trail
use eRealms::Crypto;						# Cryptographic
use eRealms::Option;						# Options handling
use eRealms::Table;						# Table Creation
use eRealms::Template;						# Template handler

#################################################################################
# Sub-Routines
#################################################################################

#####################
# Constructor
#
# Allows to class to be exported and thus spawn objects as a result

sub new
{
	#####################
	# Data members

        my $class    = shift;					# Reference to self
	my $database = shift;					# Database handle
	my $cgi      = shift;					# CGI handle
	my $param    = shift;					# Parameter hash
	my $template = shift;					# Optional template hash

        my $addr;                                               # Captured address
        my $pked;                                               # Packed address
	my $self;						# Self referncing link


	#####################
	# Program area

	# Capture address
	$addr = $cgi->remote_host();

	# Invalid IP address
	if ($addr !~ /\d{1,3}\.\d{1,3}\.\d{1,3}/)
	{
		# Socket library
		use Socket;

		# Retrieve address
		$pked = gethostbyname($addr);

		# Form into human readable
		$addr = inet_ntoa($pked);
	}

	# Define parameters
        $self  = {
		ADDR   => $addr,
		CGI    => $cgi,
		HANDLE => $database,
		PARAM  => $param, 
		TEMPL  => $template 
	 };

        # Bless $self into designated class
        bless $self, $class;

        # Return self
        return $self;
}



#####################
# Authenthication
#
# This is the primary component of this module, and this would be the 
# authenthication script, which will verify usernames and passwords.

sub Authenthication
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice of the matter

	my $addr     = $self->{'ADDR'};				# Hostname of user
	my $audit;						# Auditing entry
	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $check;						# Check registration	
	my $crypt;						# Crypted password
	my $curr;						# Current address
	my $query;						# Query to ask
	my $user;						# Users login name

	my %login;						# Login hash


	#####################
	# Program area

        # Session timeouts
	if (defined($param->{'MORI'}))
	{
		# Longer elapsed then it should be
		if ((time - $param->{'MORI'}) > 3600)
		{
			# Return false
			return 0;
		}

		# Address verification
		if (defined($param->{'ADDR'}))
		{
			# Determine address
			$curr =  $addr;
			$curr =~ s/^(\d{1,3}\.\d{1,3}).*/$1/;

			# No match
			unless ($param->{'ADDR'} =~ /^$curr/)
			{
				# Audit entry
				$audit = new eRealms::Audit($database, $cgi, $param);
				$audit->Fail($param->{'ADDR'});

				# Pre-emptive cleanup
				$param->Cleanup('ADDR');


				# Return false
				return 0;
			}
		}


		# Cleanup
		$param->Cleanup(
			'ADDR',
			'MORI'
		 );
	}


	# Build query
	# Administrative
	if ($choice == 1)
	{ 
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)
			  AND PuppeteerAdminPrivs='yes'"; 
	}

	# Supervisory
	elsif ($choice == 2)
	{ 
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword 
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)
			  AND (PuppeteerSuperPrivs='yes'
			   OR PuppeteerModPrivs='yes')";
	}

	# Non-authoritative
	elsif (($choice >= 99) || ($choice == 3))
	{ 
		$query = "SELECT PuppeteerLogin, 
			  PuppeteerPassword 
			FROM Puppeteer 
			WHERE LOWER(PuppeteerLogin)=LOWER(?)"; 
	}

	# Retrieve information
	($user, $crypt) = $database->DataGet($query, $param->{'USER'});


	# Compare passwords
	if (defined($crypt))
	{
		# If crypt
		if (defined($param->{'CRYPT'})) 
		{
			# Failed authenthication
		 	if ($param->{'CRYPT'} ne $crypt)
			{
				# Audit entry
				$audit = new eRealms::Audit($database, $cgi, $param);
				$audit->Pass($param->{'ADDR'});

				# Authenthiation failed
				return 0; 
			}
		}

		# If first time
		elsif (defined($param->{'PASS'}))
		{
			# Fail on non-match
			if (crypt(substr($param->{'PASS'}, 0, 10), $crypt) ne $crypt)
			{
				# Audit entry
				$audit = new eRealms::Audit($database, $cgi, $param);
				$audit->Pass($param->{'ADDR'});

				# Authenthication Failed
				return 0;
			}

			# Otherwise assign crypt
			else
			{ $param->{'CRYPT'} = (crypt(substr($param->{'PASS'}, 0, 10), $crypt)); }
		}

		# Assume fault
		else
		{ return 0; }
	}

	# Assume fault
	else
	{ return 0; }


	# Verify for incomplete registration
	($check) = $database->DataGetPuppeteerReg($user);

	# Return false if registration is incomplete
	if (defined($check))
	{ return 0; }


	# Database updates
	# If non-authoritative
	if (defined($param->{'JAVA'}))
	{
		#####################
		# Data members
		my $radiojava;					# Formated javascript
		my $radiochat;					# Formatted chat preferences
		my $radioformat;				# Formatted format options
		my $option;					# Options handle
	

		# Link with hash
		$database->GetHashLogin(\%login);

		# Verify options received
		$option      = new eRealms::Option();
		$radiojava   = $option->FindJavascript($login{'OptJavascript'}, $param->{'JAVA'});
		$radiochat   = $option->FindChatPrefs($login{'OptChatPref'}, $param->{'CHAT'});
		$radioformat = $option->FindFormatting($login{'OptFormat'}, $param->{'FORMAT'});

		# Javascript options
		if ($radiojava ne 'default')
		{
			# Update database entry
			$database->Quick("UPDATE Puppeteer 
				SET   PuppeteerJavascript=? 
				WHERE PuppeteerLogin=?",

			 $radiojava, 
			 $user);
		}
	
		# Chatting options
		if ($radiochat ne 'default')
		{
			# Update database entry
			$database->Quick("UPDATE Puppeteer 
				SET   PuppeteerChatPref=? 
				WHERE PuppeteerLogin=?", 

			 $radiochat, 
			 $user);
		}

		# Formatting options
		if ($radioformat ne 'default')
		{
			# Update database entry
			$database->Quick("UPDATE Puppeteer 
				SET   PuppeteerFormatting=? 
				WHERE PuppeteerLogin=?",

			 $radioformat, 
			 $user);
		}
	}

	# New authenthication login
	if ((defined($param->{'PASS'}))
	 || (defined($param->{'COOKIE'})))
	{
		# Evaluation
		eval
		{
			# Update database entry update timestamp
			$database->Quick("UPDATE Puppeteer 
 			       	SET PuppeteerTimestamp=LOCALTIMESTAMP, 
       				    PuppeteerHostname=?
       				WHERE PuppeteerLogin=?",

			 $addr,
			 $user
			);

			# Record authenthication
			$audit = new eRealms::Audit($database, $cgi, $param);
			$audit->Auth($addr);


			# Commit
			$database->Commit();
		};

		# If block failed
		if ($@)
		{
			# Issue warning
			warn("Transaction aborted: $@");

			# Undo incomplete changes
			$database->Rollback;
		}
	} 

	# Cleanup
	$param->Cleanup('CHAT', 
		'COOKIE', 
		'FORMAT', 
		'JAVA', 
		'PASS', 
	 );

	# Final assignmnet
	$param->{'USER'} = $user;


	# Return true
	return 1;
}


####################
# Check Legality
#
# Determins the legality of the user using the currently selected realm.  This 
# will also verify if a realm and puppet were selected.    

sub CheckLegality
{
	####################
	# Data members	
	my $self     = shift;					# Reference to self
	my $realm    = shift;					# Realm list

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $table;						# Table handler

	my $access;						# Realm access
	my $pass;						# Realm password
	my $verdict;						# To be or not to be
	my $verified;						# Puppet/Realm verification 

	my $count;						# Counter
	my $message;						# Warning message
	my $redirect;						# Redirect information

	####################
	# Program area

	# Verification of needed information
	if (defined($param->{'ROOM'}))
	{
		# Verification of access
		# Realm
		($access, $verified) = $database->DataGetRealmAccess($param->{'ROOM'});
		($pass) = $database->DataGetRealmPassword($param->{'ROOM'});


		# Puppet default
		unless (defined($param->{'CHAR'}))
		{
			# Puppet default retreival
			($param->{'CHAR'}) = $database->DataGetDefault(
				$param->{'USER'}, 
				$param->{'ROOM'}, 
				$access
			 );
		}

		# If puppet is defined
		if (defined($param->{'CHAR'}))
		{
			# Users
			($verdict) = $database->DataGetBannedStatus($param->{'USER'}, $param->{'ROOM'});


			# Password verification
			if (defined($pass))
			{
				# Initialize counter
				$count = 0;

				# Ecrypted passwords
				if ((defined($param->{'RCRYPT'}))
				 && ($param->{'RCRYPT'} eq $pass))
				{
					# Increment
					$count++;
				}

				# Plain text
				elsif ((defined($param->{'RPASS'})) 
				 && (crypt($param->{'RPASS'}, $pass) eq $pass))
				{
					# Increment
					$count++;


					# Create encrypted counterpart
					$param->{'RCRYPT'} = crypt($param->{'RPASS'}, $pass);

					# Cleanup
					$param->Cleanup('RPASS');
				}

				# Invalid password
				else
				{
					# Assign message
					$message = $realm->{'MsgPassword'}; 
				}
			}
			else
			{
				# Initialize counter
				$count = 1;
			}


			# Verified realms
			# Realm verification of users
			if ($verified eq 'yes')
			{
				# Pull privilege
				($verified) = $database->DataGetVerified($param->{'USER'});

				# Verification
				# Verified user
				if ($verified eq 'yes')
				{
					# Increment for allowance
					$count++;
				}

				# Tsk tsk
				else
				{
					# Assign error message
					$message = $realm->{'MsgVerified'};
				}
			}
			else
			{
				# Assume success
				$count++;
			}

		
			# Access control
			# Relaxed/Restricted realms
			if (($access ne 'list')
			 && (!defined($verdict)))
			{ $count++; }

			# List based access/canon realms
			elsif (($access eq 'list')
			 && (defined($verdict)))
			{ $count++; }

			# Banned
			else
			{
		 		# Warn of being banned
				$message = $realm->{'MsgBanned'}; 
			}

	
			# Verify for complete check
			if ($count > 2)
			{
				# Return successful
				return 1;
			}
		}
	}

	# Default error message
	$message = (defined($message)) ? $message : $realm->{'MsgInvalid'};

	# Redirection
	# Retreive inline parameters
	$redirect = $param->EmbedInline(USER=>$param->{'USER'}, CRYPT=>$param->{'CRYPT'});


	# File header
	print $cgi->header(-type=>'text/html', -refresh=>"35;URL=$realm->{'LnkPortal'}$redirect");


	# Create instance
	$table = new eRealms::Table($cgi);

	# HTML Display
	# Header
	$database->DocumentGetHeader();

	# Message Display
	$table->MakeTop();
	$table->MakeSingle($message);
	$table->MakeBlank();
	$table->MakeSingle("<A HREF=\"$realm->{LnkPortal}$redirect\"><CENTER>$realm->{TagLnkPortal}</CENTER></A>");
	$table->MakeBottom();

	# Footer
	$database->DocumentGetFooter();

	# Premature exit
	exit;							# Premature exit
}


#####################
# GatherInfo
#
# Sub-routine that will handle the display of information, comes in two flavors
# adminstrative and non-authoritative login.   Selected though alias.

sub GatherInfo
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice on the matter

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list
	my $template = $self->{'TEMPL'};			# Alternate template hash

	my $table;						# Table connection

	my %login;						# Login hash


	#####################
	# Program area

	# Create instance
	$table    = new eRealms::Table($cgi);

	# Link with hash
	$database->GetHashLogin(\%login);

	# Define Template instance
	$template = (defined($template)) ? $template : \%login;
	$tmpl     = new eRealms::Template($template);


	# Pages lacking headers
	if ($choice > 99)
	{
		print $cgi->header();
		$database->DocumentGetHeader();
	}

	# Cleanup
	$param->Cleanup('CRYPT','PASS','USER','MORI','ADDR');


	# Print titles
	# Non-authoritative
	if ($choice > 2)
	{ 
		# Display title and option to register
		$tmpl->ShowMain($login{'TagTitleGeneral'}); 
                $tmpl->Show('TmplRegister', MLINK=>"$login{LnkRegister}");
	}

	# Adminstrative
 	elsif ($choice == 1)
	{ $tmpl->ShowMain($login{'TagTitleAdmin'}); }

	# Supervisory
	elsif ($choice == 2)
	{ $tmpl->ShowMain($login{'TagTitleSuper'}); }


	# Start from and embed
	print $cgi->start_form();
	print $param->EmbedNormal($param->Flat(), AUTH=>'True'), "\n";


	# Top of table
	$table->MakeTop();

	# Username and password
	unless (defined($param->{'AUTH'}))
	{
		# Display widgets
		$table->MakeValid(
		 $tmpl->PassSub($login{'TagUsername'}), 
		 $cgi->textfield('USER', '', 15, 10));
		$table->MakeValid(
		 $tmpl->PassSub($login{'TagPassword'}), 
		 $cgi->password_field('PASS', '', 15, 15));
	}
	else
	{
		# Display error message
		$table->MakeSingle($login{'ErrLogin'});
		$table->MakeBlank();

		# Display widgets
		$table->MakeInvalid(
		 $tmpl->PassSub($login{'TagUsername'}),
		 $cgi->textfield('USER', '', 15, 10));
		$table->MakeInvalid(
		 $tmpl->PassSub($login{'TagPassword'}),
		 $cgi->password_field('PASS', '', 15, 15));
	}

	# Spacer
	$table->MakeBlank();

	# Submit and reset buttons
	$table->MakeValid($cgi->submit($login{'TxtSubmitButton'}), $cgi->reset($login{'TxtResetButton'}));
	$table->MakeBottom();


	# If non admistrative
	if ($choice == 99)
	{
		#####################
		# Data members	
		my $option;					# Option handler
		my $radiojava;					# Formated javascript
		my $radiochat;					# Formatted chat preferences
		my $radioformat;				# Formatted format options

		my @radiojava;					# Javascript options
		my @radiochat;					# Chat options
		my @radioformat;				# Formatting options


		# Create instance
		$option = new eRealms::Option();

		# Split data into arrays
		$option->Split(\@radiojava, $login{'OptJavascript'});
		$option->Split(\@radiochat, $login{'OptChatPref'});
		$option->Split(\@radioformat, $login{'OptFormat'});

		# Format options nicely
		$radiojava   = $cgi->radio_group('JAVA',   \@radiojava,   $radiojava[2],   'true');
		$radiochat   = $cgi->radio_group('CHAT',   \@radiochat,   $radiochat[4],   'true');
		$radioformat = $cgi->radio_group('FORMAT', \@radioformat, $radioformat[3], 'true');

		print "<BR>\n<BR>\n";

		# Table top and title
		$table->MakeTop();
		$table->MakeTertiary(
		 $tmpl->PassSub($login{'TagOptJavacript'}), 
		 $tmpl->PassSub($login{'TagOptChatPref'}), 
		 $tmpl->PassSub($login{'TagOptFormat'}));

		# Options and table bottom
		$table->MakeTertiary(
		 $radiojava, 
		 $radiochat, 
		 $radioformat);
		$table->MakeBottom();
	}

	# End of form
	print $cgi->end_form(), "\n";

	# If lacking information
	if ($choice > 99)
	{
		$database->DocumentGetFooter();
	}

}



#####################
# Get Verification
#
# Primary entrance which will lead to GatherInfo or Authenthication depending
# on received parameters.

sub GetVerification
{
	#####################
	# Data members
	my $self     = shift;					# Reference to self
	my $choice   = shift;					# Choice on the matter

	my $cgi      = $self->{'CGI'};				# Common gateway interface
	my $database = $self->{'HANDLE'};			# Database handle
	my $param    = $self->{'PARAM'};			# Parameter list

	my $cookie;						# System cookie
	my $crypto;						# Crypto handler
	my $table;						# Table handler

	#####################
	# Program area

	# Create instrance
	$table    = new eRealms::Table($cgi);
	
	# Verify is necessary components are there
	if ((defined($param->{'USER'})) && ((defined($param->{'PASS'})) || (defined($param->{'CRYPT'}))))
	{
		# Determine and return
		return 1 if ($self->Authenthication($choice))
	}


	# Retrieve cookie
	$cookie = $cgi->cookie(-name=>'ETHEREAL');

	# Cookie exists
	if (defined($cookie))
	{
		# Retreive infromation
		$crypto = new eRealms::Crypto();
		$crypto->DecryptParam($cookie, $param);

		# Repeat test
		if ((defined($param->{'USER'})) && (defined($param->{'CRYPT'})))
		{
			# Assign check value
			$param->{'COOKIE'} = $param->{'USER'};

			# Verify authenthication
			return 1 if ($self->Authenthication($choice))
		}
	}
	

	# Show login screen
	$self->GatherInfo($choice); 
	return 0;
}

# Alias for administrative
sub GetVerificationAdmin
{ GetVerification(shift, 1); }

# Alias for normal
sub GetVerificationNormal
{ GetVerification(shift, 3); }

# Alias for portal settings
sub GetVerificationPortal
{ GetVerification(shift, 99); }

# Alis for realm specific
sub GetVerificationRealm
{ GetVerification(shift, 100); }

# Alias for supervisory
sub GetVerificationSuper
{ GetVerification(shift, 2); }

1;								# Return true
